package ppt;

import java.io.*;
import java.util.*;
import java.math.*;
import java.net.*;

public class RTPStream{
	// Constants to hold bit masks for desired flags
	static final int flagAllOff = 0;  //         000...00000000 (empty mask)
	static final int flagbit1 = 1;    // 2^^0    000...00000001
	static final int flagbit2 = 2;    // 2^^1    000...00000010
	static final int flagbit3 = 4;    // 2^^2    000...00000100
	static final int flagbit4 = 8;    // 2^^3    000...00001000
	static final int flagbit5 = 16;   // 2^^4    000...00010000
	static final int flagbit6 = 32;   // 2^^5    000...00100000
	static final int flagbit7 = 64;   // 2^^6    000...01000000
	static final int flagbit8 = 128;  // 2^^7    000...10000000
	//...
	static final int flagbit31 = (int) Math.pow(2, 30);   // 2^^30
	//...

	//Variable to hold the status of all flags
	//OutputStream os;
	//InputStream is;
	int payload;
	long startTime;
	//byte[] header;
	int seqNumber;
	int ssrc;
	boolean debug = false;
	boolean isMarker=false;
	Vector recorders=new Vector();
	RTCPStream rtcp;

	public synchronized boolean addRecorder(Recorder r){
		boolean isExist=false;
		for (int x=0;x<recorders.size();x++) if(((Recorder)recorders.elementAt(x)).equals(r)) isExist=true;
		if (!isExist) recorders.add(r);
		return true;
	}

	public synchronized boolean removeRecorder(Recorder r)
	{
		for (int x=0;x<recorders.size();x++) 
		{
			if(((Recorder)recorders.elementAt(x)).equals(r)) 
			{
				recorders.remove(r);
				return true;
			}
		}
		return false;
	}


	public void isMarker(boolean mark){
		isMarker=mark;
	}

	public RTPStream(int payload){//OutputStream os, InputStream is,
		//this.os=os;
		//this.is=is;
		Random random = new Random();
		this.payload= payload;
		seqNumber = random.nextInt();
		ssrc = random.nextInt();
		//header = new byte[12];
		//for (int x=0;x<header.length;x++) header[x]=0;
		startTime=System.currentTimeMillis();
	}

	public boolean createRTCPStream(InputStream is,OutputStream os,boolean isPrint){
		rtcp=new RTCPStream(is,os,isPrint);
		rtcp.start();
		return true;
	}

	public boolean createRTCPStream(DatagramSocket dsctrl,InetAddress address,int port,boolean isPrint)
	{
		rtcp=new RTCPStream(dsctrl,address,port,isPrint);
		rtcp.start();
		return true;
	}

	public void resetTimer(){
		startTime=System.currentTimeMillis();
	}

	public byte[] getRTPHeader(){
		return getRTPHeader(System.currentTimeMillis()-startTime);
	}

	public byte[] getRTPHeader(long timestamp)
	{
		byte[] header = new byte[12];
		int verByte=0;
		int payloadByte=payload;
		//int seqTwoBytes=0;
		verByte|=flagbit8;
		if (isMarker) payloadByte|=flagbit8;
		header[0]=(byte)(0xFF & (verByte));
		header[1]=(byte)(0xFF & (payloadByte));
		header[2]=(byte)(0xFF & (seqNumber>>8));
		header[3]=(byte)(0xFF & (seqNumber));
		header[4]=(byte)(0xFF & (timestamp>>24));
		header[5]=(byte)(0xFF & (timestamp>>16));
		header[6]=(byte)(0xFF & (timestamp>>8));
		header[7]=(byte)(0xFF & (timestamp));
		header[8]=(byte)(0xFF & (ssrc>>24));
		header[9]=(byte)(0xFF & (ssrc>>16));
		header[10]=(byte)(0xFF & (ssrc>>8));
		header[11]=(byte)(0xFF & (ssrc));
		if (debug)  System.out.println("payload:"+payload+" seq:"+seqNumber+" timeStamp:"+timestamp+" ssrc:"+ssrc+" mark:"+isMarker);
		seqNumber++;
		return header;
	}

	class RTCPStream extends Thread{
		DatagramSocket ds=null;
		DataInputStream dis;
		OutputStream os;
		byte[] buffer = new byte[1024];
		byte[] lenArray = new byte[2];
		int firstByte = 0,secondByte = 0,rtpLen=0;
		boolean print;
		boolean isUDP=false;
		int ssrc=new Random().nextInt(),ssrc_ms,port;
		InetAddress address;

		public RTCPStream(InputStream is,OutputStream os,boolean isPrint){
			this.dis=new DataInputStream(new BufferedInputStream(is));
			this.os=os;
			print=isPrint;
		}

		public RTCPStream(DatagramSocket dsctrl,InetAddress address,int port,boolean isPrint){
			print=isPrint;
			this.address=address;
			this.port=port;
			ds=dsctrl;
			isUDP=true;
		}

		public synchronized void sendFIR(InetAddress address,int port){
			try
			{
				System.out.println("Sending FIR to "+address+":"+port);
				ByteArrayOutputStream baos=new ByteArrayOutputStream();
				DataOutputStream dos=new DataOutputStream(baos);
				dos.write(129); 
				dos.write(206); 
				dos.writeShort(2);
				dos.writeInt(ssrc);
				dos.writeInt(ssrc_ms);
				byte[] FIR=baos.toByteArray();
				if (isUDP)
				{
					System.out.println();
					DatagramPacket dp = new DatagramPacket(FIR,FIR.length,address,port);
					ds.send(dp);
				}
				else
				{
				
				}
			}catch(Exception e)	{
				e.printStackTrace();
			}
		}

		public void run(){
			int version,type,len;
			long[] others=new long[100];
			byte[] buffer=new byte[4];
			ByteArrayOutputStream recordBAOS= new ByteArrayOutputStream();
			DataOutputStream recordDOS = new DataOutputStream(recordBAOS);
			byte[] recordBuf=new byte[65536];
			byte[] dpBuffer=new byte[65536];
			DatagramPacket dp;// = new DatagramPacket(dpBuffer,dpBuffer.length);

			while(true)
			{
				try
				{
					if (isUDP)
					{
						dp = new DatagramPacket(dpBuffer,dpBuffer.length);
						ds.receive(dp);
						dis = new DataInputStream(new ByteArrayInputStream(dp.getData(),dp.getOffset(),dp.getLength()));
					}
					version = dis.readUnsignedByte();
					type = dis.readUnsignedByte();
					len = dis.readUnsignedShort();
					if (recorders.size()>0){
						recordBAOS.reset();
						dis.mark(len*4);
						recordDOS.write(version);
						recordDOS.write(type);
						recordDOS.writeShort(len);
						dis.readFully(recordBuf,0,len*4);
						recordDOS.write(recordBuf,0,len*4);
						for (int x=0;x<recorders.size();x++) ((Recorder)(recorders.elementAt(x))).write(recordBAOS.toByteArray(),false);
						dis.reset();
					}
					if (print) System.out.println("RTCP packet: "+version+" "+type+" "+len);
					/*if (rtcp[1]==202){
						if(print)
						{
							for (int x=0;x<rtcp[0]-128;x++)
							{
								dis.read(buffer,0,4);
								long ssrc=new BigInteger(buffer).longValue();
								int index=0;
								System.out.print("ssrc:"+ssrc+" ");
								int type=dis.readUnsignedByte();
								while(type!=0)
								{
									//index=0;
									System.out.print("type:"+type+" ");
									int len=dis.readUnsignedByte();
									index+=len+2;
									System.out.print("len:"+len+" ");
									byte[] text=new byte[len];
									dis.read(text,0,len);
									System.out.print("text:"+new String(text)+" ");
									type=dis.readUnsignedByte();
								}
								System.out.println("read"+index+" "+(4-(index+1)%4)%4);
								dis.read(buffer,0,(4-(index+1)%4)%4);
							}
							System.out.println();
						}
						else{
							for (int x=0;x<rtcp[2];x++) 
							{
								dis.read(buffer,0,4);
								others[x]=new BigInteger(buffer).longValue();
							}	
						}	
					}*/
					//else{
					for (int x=0;x<len;x++) 
					{
						dis.read(buffer,0,4);
						others[x]=new BigInteger(1,buffer).longValue();
					}
					if (type==200) ssrc_ms=(int)others[0];
					if (print) 
					{
						for (int x=0;x<len;x++) System.out.print(others[x]+" ");
						System.out.println();
					}
					//}
				}
				catch(Exception e){
					System.out.println("rtcpStream Error:"+e);
					e.printStackTrace();
					break;
				}

			}
		}
	
	}

}
