package ppt;

import java.io.*;
import java.util.*;
import java.math.*;
import java.net.*;

public class RTPStream{
	// Constants to hold bit masks for desired flags
	static final int flagAllOff = 0;  //         000...00000000 (empty mask)
	static final int flagbit1 = 1;    // 2^^0    000...00000001
	static final int flagbit2 = 2;    // 2^^1    000...00000010
	static final int flagbit3 = 4;    // 2^^2    000...00000100
	static final int flagbit4 = 8;    // 2^^3    000...00001000
	static final int flagbit5 = 16;   // 2^^4    000...00010000
	static final int flagbit6 = 32;   // 2^^5    000...00100000
	static final int flagbit7 = 64;   // 2^^6    000...01000000
	static final int flagbit8 = 128;  // 2^^7    000...10000000
									  //...
	static final int flagbit31 = (int) Math.pow(2, 30);   // 2^^30
														  //...
	
	static final int TCP_MODE = 1;
	static final int UDP_MODE = 2;
	
	static final int NO_RELIABILITY   = 0;
	static final int ONE_NACK         = 1;
	static final int FULL_RELIABILITY = 2;
	int reliability = NO_RELIABILITY;
	
	//Variable to hold the status of all flags
	//OutputStream os;
	//InputStream is;
	int payload;
	long startTime;
	int rtp_mode, rtcp_mode;
	//byte[] header;
	int seqNumber;
	int ssrc;
	boolean debug = false;
	boolean useRTPFraming = false;
	//boolean reliableComm = false;
	RTPMessage[] reliableBuffer;
	long[] NACKSendTimes;
	int reliableTimeOutMillis = 1000;
	int maximumRetryNumber = 3;
	int reliableBufferSize = 5000;
	int reliableBufferHead = 0, reliableBufferEnd = 0;
	long reliableNACKTime;
	int lastRTPSequenceNum=-1;
	boolean isMarker=false;
	Vector recorders=new Vector();
	
	Socket rtpTcp, rtcpTcp;
	DatagramSocket rtpUdp, rtcpUdp;
	DataInputStream dis;
	DataOutputStream dos;
	RTCPStream rtcp;
	RTPMessage rtpMes;
	
	InetAddress remoteAddress;
	long lastNackTime=0;
	long totalPackets,lostPackets,dupPackets,initialNACKs,dupNACKs;
	long ltotalPackets,llostPackets,ldupPackets,linitialNACKs,ldupNACKs;
	
	byte[] dpBuffer = new byte[65536];
	
	int dl =2;
	
	
	public synchronized boolean addRecorder(Recorder r){
		boolean isExist=false;
		for (int x=0;x<recorders.size();x++) if(((Recorder)recorders.elementAt(x)).equals(r)) isExist=true;
		if (!isExist) recorders.add(r);
		return true;
	}
	
	public synchronized boolean removeRecorder(Recorder r)
	{
		for (int x=0;x<recorders.size();x++) 
		{
			if(((Recorder)recorders.elementAt(x)).equals(r)) 
			{
				recorders.remove(r);
				return true;
			}
		}
		return false;
	}
	
	
	public void isMarker(boolean mark){
		isMarker=mark;
	}
	
	public InetAddress getRemoteAddress(){
		return remoteAddress;
	}
	
	public boolean setReliabilityLevel(int x){
		if (!(x==NO_RELIABILITY || x==ONE_NACK || x== FULL_RELIABILITY)) {
			//System.out.println("Unidentified Reliability Level:"+x);
			return false;
		}
		reliability=x;
		if (x!=NO_RELIABILITY){
			reliableBuffer = new RTPMessage[reliableBufferSize];
			NACKSendTimes= new long[reliableBufferSize];
		}
		if (x==FULL_RELIABILITY){
			reliableTimeOutMillis*=5;
		}
		return true;
	}
	
	public String getStats(){
		String lastSec= "RTP Statistics Sec total:"+(totalPackets-ltotalPackets)+" lost:"+(lostPackets-llostPackets)+" dup:"+(dupPackets-ldupPackets)+" iNack:"+(initialNACKs-linitialNACKs)+" dNACK:"+(dupNACKs-ldupNACKs);
		String s      = "\nRTP Statistics All total:"+totalPackets+" lost:"+lostPackets+" dup:"+dupPackets+" iNack:"+initialNACKs+" dNACK:"+dupNACKs;
		ltotalPackets=totalPackets; llostPackets=lostPackets; ldupPackets=dupPackets; linitialNACKs=initialNACKs; ldupNACKs=dupNACKs;
		return lastSec+s;
	}
	
	public RTPMessage read() {
		//if (dl>1)  System.out.print(" *1* ");
		boolean isContinue = true;
		long longTemp;
		
		RTPMessage last;
		if (reliability==NO_RELIABILITY) {
			last = readRTPMessage(0);
			if (lastRTPSequenceNum == -1) lastRTPSequenceNum = last.seqNumber - 1;
			if (last.seqNumber>lastRTPSequenceNum||(lastRTPSequenceNum>65000&&last.seqNumber<1000)) {
				lastRTPSequenceNum = last.seqNumber;
				//if (dl>1)  System.out.print(" *4-"+lastRTPSequenceNum+"* ");
				return last;
			}
			else if (dl>1) System.out.println("DISCARD "+last.seqNumber);
		}
		else {
			while (isContinue)			{
				if ((reliableBufferEnd - reliableBufferHead)%reliableBufferSize == 0)
				{
					//if (dl > 1) System.out.print(" *2* ");
					last = readRTPMessage(0);
					if (lastRTPSequenceNum == -1) lastRTPSequenceNum = last.seqNumber - 1;
					if (last.seqNumber == ((lastRTPSequenceNum + 1) % 65536))
					{
						lastRTPSequenceNum = last.seqNumber;
						//if (dl > 1) System.out.print(" *3-" + last.seqNumber+"* ");
						return last;
					}
					else if (last.seqNumber==lastRTPSequenceNum) {dupPackets++; 
					//System.out.println("DISCARD duplicate");
					} //DISCARD
					else
					{
						//if (dl > 1) System.out.println("the packet is out of order seq_num: " + last.seqNumber + " lastRcvdSeqNum: " + lastRTPSequenceNum);
						int diff = last.seqNumber - lastRTPSequenceNum;
						if (diff<0) {
								if (diff>-5000) dupPackets++;
								diff+=65536;
							}
						//if (diff > 5000) System.out.println("^^^ DISCARD "+last.seqNumber+" ^^^"); //DISCARD
						else
						{
							int pos = (reliableBufferHead + diff - 2) % reliableBufferSize;
							//System.out.println("pos:"+posOri+" "+pos);
							reliableBuffer[pos] = last;
							reliableBufferEnd = (pos+ 1)  ;
							if (dl > 0) System.out.println("packet inserted to pos:" + pos + " BufBegin" + reliableBufferHead + " BufEnd" + reliableBufferEnd+" nack:"
															+(lastRTPSequenceNum + 1)+"-"+last.seqNumber);
							for (int x = lastRTPSequenceNum + 1; x < last.seqNumber; x=(x+1)%65536){ rtcp.sendNACK(x%65536);initialNACKs++;}
							
							
							for (int x=reliableBufferHead;x!=(reliableBufferEnd)%reliableBufferSize;x=(x+1)%reliableBufferSize) NACKSendTimes[x] = System.currentTimeMillis();
							//if (dl>1)System.out.println("xxxxxxx"+NACKSendTimes[reliableBufferHead]);
							lastNackTime=System.currentTimeMillis();
						}
					}
				}
				else
				{
					int startingPos = reliableBufferHead;
					while (reliableBuffer[startingPos] == null) startingPos=(startingPos+1)% reliableBufferSize;
					if (System.currentTimeMillis() - NACKSendTimes[startingPos] > reliableTimeOutMillis) reliableBufferHead=startingPos;
					
					if ( (reliableBuffer[reliableBufferHead] != null) && ( (reliableBuffer[reliableBufferHead].seqNumber == (lastRTPSequenceNum + 1) % 65536)
																		   || (System.currentTimeMillis() - NACKSendTimes[reliableBufferHead] > reliableTimeOutMillis) )
						 )
					{
						if (dl > 0 && (reliableBuffer[reliableBufferHead].seqNumber != (lastRTPSequenceNum + 1) % 65536)) {
									System.out.println("RETURNING head ind:"+reliableBufferHead+" seq:" + reliableBuffer[reliableBufferHead].seqNumber+ " "+
														" "+ lastRTPSequenceNum+" len:" + reliableBuffer[reliableBufferHead].length+" "+
														(System.currentTimeMillis()-NACKSendTimes[reliableBufferHead])+"ms");
									lostPackets+=reliableBuffer[reliableBufferHead].seqNumber-lastRTPSequenceNum;
						}
					//	else if(dl>0) System.out.print("*5-"+reliableBuffer[reliableBufferHead].seqNumber+"*");
						RTPMessage temp = reliableBuffer[reliableBufferHead];
						reliableBuffer[reliableBufferHead]=null;
						reliableBufferHead = ++reliableBufferHead % reliableBufferSize;
						lastRTPSequenceNum = temp.seqNumber;
						return temp;
					}
					else {
						last = readRTPMessage(500);
						if (last != null) {
							if (last.seqNumber == (lastRTPSequenceNum + 1) % 65536){
								lastRTPSequenceNum = last.seqNumber;
								//if (dl > 1) System.out.println("\nloop-the packet is in order seq_num:" + last.seqNumber+" "+(System.currentTimeMillis()-NACKSendTimes[reliableBufferHead])+"ms");
								if (reliableBuffer[reliableBufferHead]==null) reliableBufferHead=++reliableBufferHead%reliableBufferSize;
								return last;
							}
							
							if (System.currentTimeMillis()-lastNackTime>500){
								//if (dl > 0) System.out.println("read returned "+last.seqNumber+" sending NACK for "+(lastRTPSequenceNum+1)+" "+System.currentTimeMillis());
								rtcp.sendNACK( (lastRTPSequenceNum + 1) % 65536);
								int index = reliableBufferHead;
								
								while (reliableBuffer[index] == null) {
									rtcp.sendNACK( (lastRTPSequenceNum+2+index-reliableBufferHead) % 65536);
									dupNACKs++;
									//NACKSendTimes[index] = System.currentTimeMillis();
									index=++index% reliableBufferSize;
								}
								
								lastNackTime=System.currentTimeMillis();
							}
							if (last.seqNumber==lastRTPSequenceNum) {dupPackets++; 
							//System.out.println("DISCARD"); 
							continue; }//DISCARD
							
						//if (dl > 1) System.out.print(" loop-OOO : " + last.seqNumber + " lastRcvdSeqNum:" + lastRTPSequenceNum+" ");
							int diff = last.seqNumber - lastRTPSequenceNum;
							if (diff<0) {
								if (diff>-5000) dupPackets++;
								diff+=65536;
							}
						//	if (diff > 5000)  System.out.println("^^^ DISCARD "+last.seqNumber+" ^^^"); //DISCARD
							else{
								int pos=(reliableBufferHead + diff - 2) % reliableBufferSize;
								reliableBuffer[pos] = last;
								//if (dl > 2) System.out.println("packet inserted to pos:"+pos+" BufBegin"+reliableBufferHead+" BufEnd"+reliableBufferEnd);
								int diff2=last.seqNumber-reliableBuffer[(reliableBufferEnd-1)].seqNumber; 
								if (diff2<0) diff2+=65536;
								if ( diff2<4800 ){
									int temp = reliableBuffer[(reliableBufferEnd-1)].seqNumber;
								//	if (dl>2) System.out.println("end moved. nacK"+(temp+1)+"-"+last.seqNumber);
									if (dl>0&&(temp+1)!=last.seqNumber) System.out.println("end moved. Sending NACK:"+(temp+1)+"-"+last.seqNumber);
									for (int x = temp+1; x < last.seqNumber; x=(x+1)%65536) { rtcp.sendNACK(x%65536);initialNACKs++;}
								    

									for (int x=(reliableBufferEnd-1);x!=(pos+1)%reliableBufferSize;x=(x+1)%reliableBufferSize) NACKSendTimes[x] = System.currentTimeMillis();
									
									reliableBufferEnd = (pos+1) ;
									//if (reliableBufferEnd==%reliableBufferSize+1) 
								}
								//else if (dl>1) System.out.println("--- RETRANSMISSION-"+last.seqNumber+"---");
								
							}
							
						}
						else {
							if (dl > 0) System.out.println("read returned null sending NACK for "+(lastRTPSequenceNum+1)+" "+System.currentTimeMillis());
							rtcp.sendNACK( (lastRTPSequenceNum + 1) % 65536);
							int index = reliableBufferHead;
								while (reliableBuffer[index] == null) {
									rtcp.sendNACK( (lastRTPSequenceNum+2+index-reliableBufferHead) % 65536);
									dupNACKs++;
									//NACKSendTimes[index] = System.currentTimeMillis();
									index=(index+1)% reliableBufferSize;
								}
								
							lastNackTime=System.currentTimeMillis();
						}
					}
				}
			}//while(true)
		}
		return null;
	}
	
	public RTPMessage readRTPMessage(int timeout) {
		int rtpLen;
		try{
			if (rtp_mode == TCP_MODE)
			{
				rtpLen = dis.readUnsignedShort();
			}
			else {
				DatagramPacket dp = new DatagramPacket(dpBuffer, dpBuffer.length);
				rtpUdp.setSoTimeout(timeout);
				rtpUdp.receive(dp);
				dis = new DataInputStream(new ByteArrayInputStream(dp.getData(), dp.getOffset(), dp.getLength()));
				if (remoteAddress==null) {
					remoteAddress=dp.getAddress(); //port=dp.getPort();
				}
				//System.out.println("len:"+dp.getLength());
				rtpLen = dp.getLength();
			}
			byte[] rtpHeader = new byte[12];
			dis.readFully(rtpHeader);
			byte[] dataX = new byte[rtpLen-12];
			dis.readFully(dataX);
			rtpMes = new RTPMessage(rtpHeader,dataX);
			totalPackets++;
			return rtpMes;
		}catch(SocketTimeoutException ex){
			//ex.printStackTrace();
		} catch(Exception e){
			e.printStackTrace();
		}
		return null;
	}
	
	public void setSockets(Socket rtp, Socket rtcp) {
		rtp_mode = TCP_MODE;
		rtcp_mode = TCP_MODE;
		rtpTcp = rtp;
		rtcpTcp = rtcp;
		try{
			int size1 = rtpTcp.getReceiveBufferSize();
			rtpTcp.setReceiveBufferSize(256 * 1024);
			
			//System.out.println("Receive Buffer:" + size1 + " " + rtpTcp.getReceiveBufferSize());
			rtpTcp.setTcpNoDelay(true);
			rtpTcp.setKeepAlive(true);
			//rtcpTcp.setKeepAlive(true);
			
			dis = new DataInputStream(new BufferedInputStream(rtp.getInputStream()));
			dos = new DataOutputStream(new BufferedOutputStream(rtp.getOutputStream()));
		}catch(Exception e){
			e.printStackTrace();
		}
		try{
		//	createRTCPStream(rtcp.getInputStream(),rtcp.getOutputStream(),false);
		}catch(Exception e){
			e.printStackTrace();
		}
	}
	
	public void setSockets(DatagramSocket rtp, DatagramSocket rtcp) { 
		
	}
	
	public void setRemoteAddress(InetAddress address,int port,int mode) {
		try{
			if (address.isMulticastAddress())
			{
				MulticastSocket msUdp = new MulticastSocket(port);
				msUdp.joinGroup(address);
				msUdp.setLoopbackMode(true);
				msUdp.setTimeToLive(255);
				//ms.connect(address,port);
				MulticastSocket mscUdp = new MulticastSocket(port + 1);
				mscUdp.joinGroup(address);
				mscUdp.setLoopbackMode(true);
				mscUdp.setTimeToLive(255);
				//msctrl.connect(address,port);
				rtp_mode = UDP_MODE;
				rtcp_mode = UDP_MODE;
				rtpUdp = msUdp;
				rtcpUdp = mscUdp;
				//System.out.println(rtcpUdp+"");
				createRTCPStream(rtcpUdp, address, port + 1, false);
				
				System.out.println("Joined to Multicast Group " + address + ".");
				System.out.println("LB Modes " + msUdp.getLoopbackMode() + " " + mscUdp.getLoopbackMode());
			}
			else if (mode==TCP_MODE)
			{
				rtpTcp = new Socket();
				rtpTcp.connect(new InetSocketAddress(address, port));
			//	rtcpTcp = new Socket(address, port + 1);
				
				setSockets(rtpTcp, rtcpTcp);
			}
			else if (mode == UDP_MODE) {
				rtpUdp = new DatagramSocket(port);
				rtcpUdp = new DatagramSocket(port+1);
				rtp_mode = UDP_MODE;
				rtcp_mode = UDP_MODE;
			}
		}
		catch (Exception e)
	{
			e.printStackTrace();
	}
	}
	
	public void enable_rtp_framing(boolean x) {
		useRTPFraming = x;
	}
	
	public RTPStream(int payload){//OutputStream os, InputStream is,
								  //this.os=os;
								  //this.is=is;
		Random random = new Random();
		this.payload= payload;
		seqNumber = random.nextInt();
		ssrc = random.nextInt();
		//header = new byte[12];
		//for (int x=0;x<header.length;x++) header[x]=0;
		startTime=System.currentTimeMillis();
	}
	
	public boolean createRTCPStream(InputStream is,OutputStream os,boolean isPrint){
		rtcp=new RTCPStream(is,os,isPrint);
		rtcp.start();
		return true;
	}
	
	public boolean createRTCPStream(DatagramSocket dsctrl,InetAddress address,int port,boolean isPrint)
	{
		rtcp=new RTCPStream(dsctrl,address,port,isPrint);
		rtcp.start();
		return true;
	}
	
	public void resetTimer(){
		startTime=System.currentTimeMillis();
	}
	
	public byte[] getRTPHeader(){
		return getRTPHeader(System.currentTimeMillis()-startTime);
	}
	
	public byte[] getRTPHeader(long timestamp)
	{
		byte[] header = new byte[12];
		int verByte=0;
		int payloadByte=payload;
		//int seqTwoBytes=0;
		verByte|=flagbit8;
		if (isMarker) payloadByte|=flagbit8;
		header[0]=(byte)(0xFF & (verByte));
		header[1]=(byte)(0xFF & (payloadByte));
		header[2]=(byte)(0xFF & (seqNumber>>8));
		header[3]=(byte)(0xFF & (seqNumber));
		header[4]=(byte)(0xFF & (timestamp>>24));
		header[5]=(byte)(0xFF & (timestamp>>16));
		header[6]=(byte)(0xFF & (timestamp>>8));
		header[7]=(byte)(0xFF & (timestamp));
		header[8]=(byte)(0xFF & (ssrc>>24));
		header[9]=(byte)(0xFF & (ssrc>>16));
		header[10]=(byte)(0xFF & (ssrc>>8));
		header[11]=(byte)(0xFF & (ssrc));
		if (debug)  System.out.println("payload:"+payload+" seq:"+seqNumber+" timeStamp:"+timestamp+" ssrc:"+ssrc+" mark:"+isMarker);
	//	System.out.println("payload:"+payload+" seq:"+seqNumber+" timeStamp:"+timestamp+" ssrc:"+ssrc+" mark:"+isMarker);
		seqNumber++;
		return header;
	}
	
	class RTCPStream extends Thread{
		DatagramSocket ds=null;
		DataInputStream dis;
		OutputStream os;
		byte[] buffer = new byte[1024];
		byte[] lenArray = new byte[2];
		int firstByte = 0,secondByte = 0,rtpLen=0;
		boolean print;
		boolean isUDP=false;
		int ssrc=new Random().nextInt(),ssrc_ms,port;
		InetAddress address;
		
		public RTCPStream(InputStream is,OutputStream os,boolean isPrint){
			this.dis=new DataInputStream(new BufferedInputStream(is));
			this.os=os;
			print=isPrint;
		}
		
		public RTCPStream(DatagramSocket dsctrl,InetAddress address,int port,boolean isPrint){
			print=isPrint;
			//this.address=address;
			this.port=port;
			ds=dsctrl;
			//System.out.println(ds+"");
			isUDP=true;
		}
		
		public synchronized void sendNACK(int seq_num){
			try
		{
			if (dl>1) System.out.print(" NACK "+seq_num+" ");
			ByteArrayOutputStream baos=new ByteArrayOutputStream();
			DataOutputStream dos=new DataOutputStream(baos);
			dos.write(129); 
			dos.write(206); 
			dos.writeShort(1);
			dos.writeInt(seq_num);
			//dos.writeInt(ssrc_ms);
			byte[] NACK=baos.toByteArray();
			if (isUDP)
			{
				//System.out.println();
				DatagramPacket dp = new DatagramPacket(NACK,NACK.length,address,port);
				ds.send(dp);
			}
			else
			{
				
			}
		}catch(Exception e)	{
			e.printStackTrace();
		}
		}
		
		
		public synchronized void sendFIR(){//InetAddress address,int port){
			try
		{
			System.out.println("Sending FIR to "+address+":"+port);
			ByteArrayOutputStream baos=new ByteArrayOutputStream();
			DataOutputStream dos=new DataOutputStream(baos);
			dos.write(129); 
			dos.write(206); 
			dos.writeShort(2);
			dos.writeInt(ssrc);
			dos.writeInt(ssrc_ms);
			byte[] FIR=baos.toByteArray();
			if (isUDP)
			{
				System.out.println();
				DatagramPacket dp = new DatagramPacket(FIR,FIR.length,address,port);
				ds.send(dp);
			}
			else
			{
				
			}
		}catch(Exception e)	{
			e.printStackTrace();
		}
		}
		
		public void run(){
			int version,type,len;
			long[] others=new long[200];
			byte[] buffer=new byte[4];
			ByteArrayOutputStream recordBAOS= new ByteArrayOutputStream();
			DataOutputStream recordDOS = new DataOutputStream(recordBAOS);
			byte[] recordBuf=new byte[65536];
			byte[] dpBuffer=new byte[65536];
			DatagramPacket dp;// = new DatagramPacket(dpBuffer,dpBuffer.length);
				
				while(true)
				{
					try
				{
					if (isUDP)
					{
						dp = new DatagramPacket(dpBuffer,dpBuffer.length);
						ds.receive(dp);
						if (address==null) {
							address=dp.getAddress(); //port=dp.getPort();
						}
						dis = new DataInputStream(new ByteArrayInputStream(dp.getData(),dp.getOffset(),dp.getLength()));
					}
					version = dis.readUnsignedByte();
					type = dis.readUnsignedByte();
					len = dis.readUnsignedShort();
					if (recorders.size()>0){
						recordBAOS.reset();
						dis.mark(len*4);
						recordDOS.write(version);
						recordDOS.write(type);
						recordDOS.writeShort(len);
						dis.readFully(recordBuf,0,len*4);
						recordDOS.write(recordBuf,0,len*4);
						for (int x=0;x<recorders.size();x++) ((Recorder)(recorders.elementAt(x))).write(recordBAOS.toByteArray(),false);
						dis.reset();
					}
					if (print) System.out.println("RTCP packet: "+version+" "+type+" "+len);
					/*if (rtcp[1]==202){
						if(print)
					{
							for (int x=0;x<rtcp[0]-128;x++)
							{
								dis.read(buffer,0,4);
								long ssrc=new BigInteger(buffer).longValue();
								int index=0;
								System.out.print("ssrc:"+ssrc+" ");
								int type=dis.readUnsignedByte();
								while(type!=0)
								{
									//index=0;
									System.out.print("type:"+type+" ");
									int len=dis.readUnsignedByte();
									index+=len+2;
									System.out.print("len:"+len+" ");
									byte[] text=new byte[len];
									dis.read(text,0,len);
									System.out.print("text:"+new String(text)+" ");
									type=dis.readUnsignedByte();
								}
								System.out.println("read"+index+" "+(4-(index+1)%4)%4);
								dis.read(buffer,0,(4-(index+1)%4)%4);
							}
							System.out.println();
					}
					else{
						for (int x=0;x<rtcp[2];x++) 
						{
							dis.read(buffer,0,4);
							others[x]=new BigInteger(buffer).longValue();
						}	
					}	
					}*/
					//else{
					for (int x=0;x<len;x++) 
					{
						dis.read(buffer,0,4);
						others[x]=new BigInteger(1,buffer).longValue();
					}
					if (type==200) ssrc_ms=(int)others[0];
					if (print) 
					{
						for (int x=0;x<len;x++) System.out.print(others[x]+" ");
						System.out.println();
					}
					//}
				}
					catch(Exception e){
						System.out.println("rtcpStream Error:"+e);
						e.printStackTrace();
						break;
					}
					
				}
		}
		
	}
	
}
